#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © Nekokatt 2019-2020
#
# This file is part of Hikari.
#
# Hikari is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Hikari is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Hikari. If not, see <https://www.gnu.org/licenses/>.
"""Package metadata."""
__all__ = ["__author__", "__copyright__", "__email__", "__license__", "__version__", "__url__"]

__author__ = "Nekokatt"
__copyright__ = "© 2019-2020 Nekokatt"
__email__ = "3903853-nekokatt@users.noreply.gitlab.com"
__license__ = "LGPL-3.0-ONLY"
__version__ = "1.0.1.dev66"
__url__ = "https://gitlab.com/nekokatt/hikari"
